/*
 * bsc1196959_aio
 *
 * Fix for CVE-2021-39698, bsc#1196959 (fs/aio.c part)
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

static const struct file_operations (*klpe_signalfd_fops);

/* klp-ccp: from fs/aio.c */
#define pr_fmt(fmt) "%s: " fmt, __func__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/aio_abi.h>
#include <linux/export.h>
#include <linux/syscalls.h>
#include <linux/backing-dev.h>
#include <linux/refcount.h>
#include <linux/uio.h>
#include <linux/sched/signal.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/percpu.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/aio.h>
#include <linux/highmem.h>
#include <linux/workqueue.h>
#include <linux/security.h>
#include <linux/eventfd.h>

/* klp-ccp: from include/linux/eventfd.h */
#ifdef CONFIG_EVENTFD

static int (*klpe_eventfd_wake_count);

static inline bool klpr_eventfd_signal_count(void)
{
	return *this_cpu_ptr(&(*klpe_eventfd_wake_count));
}

#else /* CONFIG_EVENTFD */
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from fs/aio.c */
#include <linux/blkdev.h>
#include <linux/compat.h>
#include <linux/percpu-refcount.h>
#include <asm/kmap_types.h>
#include <linux/uaccess.h>

struct aio_ring {
	unsigned	id;	/* kernel internal index number */
	unsigned	nr;	/* number of io_events */
	unsigned	head;	/* Written to by userland or under ring_lock
				 * mutex by aio_read_events_ring(). */
	unsigned	tail;

	unsigned	magic;
	unsigned	compat_features;
	unsigned	incompat_features;
	unsigned	header_length;	/* size of aio_ring */


	struct io_event		io_events[0];
};

#define AIO_RING_PAGES	8

struct kioctx {
	struct percpu_ref	users;
	atomic_t		dead;

	struct percpu_ref	reqs;

	unsigned long		user_id;

	struct __percpu kioctx_cpu *cpu;

	/*
	 * For percpu reqs_available, number of slots we move to/from global
	 * counter at a time:
	 */
	unsigned		req_batch;
	/*
	 * This is what userspace passed to io_setup(), it's not used for
	 * anything but counting against the global max_reqs quota.
	 *
	 * The real limit is nr_events - 1, which will be larger (see
	 * aio_setup_ring())
	 */
	unsigned		max_reqs;

	/* Size of ringbuffer, in units of struct io_event */
	unsigned		nr_events;

	unsigned long		mmap_base;
	unsigned long		mmap_size;

	struct page		**ring_pages;
	long			nr_pages;

	struct rcu_work		free_rwork;	/* see free_ioctx() */

	/*
	 * signals when all in-flight requests are done
	 */
	struct ctx_rq_wait	*rq_wait;

	struct {
		/*
		 * This counts the number of available slots in the ringbuffer,
		 * so we avoid overflowing it: it's decremented (if positive)
		 * when allocating a kiocb and incremented when the resulting
		 * io_event is pulled off the ringbuffer.
		 *
		 * We batch accesses to it with a percpu version.
		 */
		atomic_t	reqs_available;
	} ____cacheline_aligned_in_smp;

	struct {
		spinlock_t	ctx_lock;
		struct list_head active_reqs;	/* used for cancellation */
	} ____cacheline_aligned_in_smp;

	struct {
		struct mutex	ring_lock;
		wait_queue_head_t wait;
	} ____cacheline_aligned_in_smp;

	struct {
		unsigned	tail;
		unsigned	completed_events;
		spinlock_t	completion_lock;
	} ____cacheline_aligned_in_smp;

	struct page		*internal_pages[AIO_RING_PAGES];
	struct file		*aio_ring_file;

	unsigned		id;
};

struct fsync_iocb {
	struct file		*file;
	struct work_struct	work;
	bool			datasync;
	struct cred		*creds;
};

struct poll_iocb {
	struct file		*file;
	struct wait_queue_head	*head;
	__poll_t		events;
	bool			done;
	bool			cancelled;
	struct wait_queue_entry	wait;
	struct work_struct	work;
};

struct aio_kiocb {
	union {
		struct file		*ki_filp;
		struct kiocb		rw;
		struct fsync_iocb	fsync;
		struct poll_iocb	poll;
	};

	struct kioctx		*ki_ctx;
	kiocb_cancel_fn		*ki_cancel;

	struct io_event		ki_res;

	struct list_head	ki_list;	/* the aio core uses this
						 * for cancellation */
	refcount_t		ki_refcnt;

	/*
	 * If the aio_resfd field of the userspace iocb is not zero,
	 * this is the underlying eventfd context to deliver events to.
	 */
	struct eventfd_ctx	*ki_eventfd;
};

static struct kmem_cache	*(*klpe_kiocb_cachep);

#define AIO_EVENTS_PER_PAGE	(PAGE_SIZE / sizeof(struct io_event))
#define AIO_EVENTS_FIRST_PAGE	((PAGE_SIZE - sizeof(struct aio_ring)) / sizeof(struct io_event))
#define AIO_EVENTS_OFFSET	(AIO_EVENTS_PER_PAGE - AIO_EVENTS_FIRST_PAGE)

static void (*klpe_refill_reqs_available)(struct kioctx *ctx, unsigned head,
                                  unsigned tail);

static inline void klpr_iocb_destroy(struct aio_kiocb *iocb)
{
	if (iocb->ki_eventfd)
		eventfd_ctx_put(iocb->ki_eventfd);
	if (iocb->ki_filp)
		fput(iocb->ki_filp);
	percpu_ref_put(&iocb->ki_ctx->reqs);
	kmem_cache_free((*klpe_kiocb_cachep), iocb);
}

static void klpr_aio_complete(struct aio_kiocb *iocb)
{
	struct kioctx	*ctx = iocb->ki_ctx;
	struct aio_ring	*ring;
	struct io_event	*ev_page, *event;
	unsigned tail, pos, head;
	unsigned long	flags;

	/*
	 * Add a completion event to the ring buffer. Must be done holding
	 * ctx->completion_lock to prevent other code from messing with the tail
	 * pointer since we might be called from irq context.
	 */
	spin_lock_irqsave(&ctx->completion_lock, flags);

	tail = ctx->tail;
	pos = tail + AIO_EVENTS_OFFSET;

	if (++tail >= ctx->nr_events)
		tail = 0;

	ev_page = kmap_atomic(ctx->ring_pages[pos / AIO_EVENTS_PER_PAGE]);
	event = ev_page + pos % AIO_EVENTS_PER_PAGE;

	*event = iocb->ki_res;

	kunmap_atomic(ev_page);
	flush_dcache_page(ctx->ring_pages[pos / AIO_EVENTS_PER_PAGE]);

	pr_debug("%p[%u]: %p: %p %Lx %Lx %Lx\n", ctx, tail, iocb,
		 (void __user *)(unsigned long)iocb->ki_res.obj,
		 iocb->ki_res.data, iocb->ki_res.res, iocb->ki_res.res2);

	/* after flagging the request as done, we
	 * must never even look at it again
	 */
	smp_wmb();	/* make event visible before updating tail */

	ctx->tail = tail;

	ring = kmap_atomic(ctx->ring_pages[0]);
	head = ring->head;
	ring->tail = tail;
	kunmap_atomic(ring);
	flush_dcache_page(ctx->ring_pages[0]);

	ctx->completed_events++;
	if (ctx->completed_events > 1)
		(*klpe_refill_reqs_available)(ctx, head, tail);
	spin_unlock_irqrestore(&ctx->completion_lock, flags);

	pr_debug("added to ring %p at [%u]\n", iocb, tail);

	/*
	 * Check if the user asked us to deliver the result through an
	 * eventfd. The eventfd_signal() function is safe to be called
	 * from IRQ context.
	 */
	if (iocb->ki_eventfd)
		eventfd_signal(iocb->ki_eventfd, 1);

	/*
	 * We have to order our ring_info tail store above and test
	 * of the wait list below outside the wait lock.  This is
	 * like in wake_up_bit() where clearing a bit has to be
	 * ordered with the unlocked test.
	 */
	smp_mb();

	if (waitqueue_active(&ctx->wait))
		wake_up(&ctx->wait);
}

static inline void klpr_iocb_put(struct aio_kiocb *iocb)
{
	if (refcount_dec_and_test(&iocb->ki_refcnt)) {
		klpr_aio_complete(iocb);
		klpr_iocb_destroy(iocb);
	}
}

static void (*klpe_aio_poll_put_work)(struct work_struct *work);

void klpp_aio_poll_complete_work(struct work_struct *work)
{
	struct poll_iocb *req = container_of(work, struct poll_iocb, work);
	struct aio_kiocb *iocb = container_of(req, struct aio_kiocb, poll);
	struct poll_table_struct pt = { ._key = req->events };
	struct kioctx *ctx = iocb->ki_ctx;
	__poll_t mask = 0;

	/*
	 * Fix CVE-2021-39698
	 *  -2 lines, +12 lines
	 */
	if (!READ_ONCE(req->cancelled)) {
		mask = vfs_poll(req->file, &pt) & req->events;

		if (req->file->f_op == &(*klpe_signalfd_fops) &&
		    iocb->ki_res.res == -ECANCELED) {
			/*
			 * A POLLFREE/EPOLLHUP event, c.f the livepatched
			 * aio_poll_wake().
			 */
			mask |= (EPOLLHUP | POLLFREE) & req->events;
		}
	}

	/*
	 * Note that ->ki_cancel callers also delete iocb from active_reqs after
	 * calling ->ki_cancel.  We need the ctx_lock roundtrip here to
	 * synchronize with them.  In the cancellation case the list_del_init
	 * itself is not actually needed, but harmless so we keep it in to
	 * avoid further branches in the fast path.
	 */
	spin_lock_irq(&ctx->ctx_lock);
	if (!mask && !READ_ONCE(req->cancelled)) {
		add_wait_queue(req->head, &req->wait);
		spin_unlock_irq(&ctx->ctx_lock);
		return;
	}
	list_del_init(&iocb->ki_list);
	iocb->ki_res.res = mangle_poll(mask);
	req->done = true;
	spin_unlock_irq(&ctx->ctx_lock);

	klpr_iocb_put(iocb);
}

int klpp_aio_poll_cancel(struct kiocb *iocb)
{
	struct aio_kiocb *aiocb = container_of(iocb, struct aio_kiocb, rw);
	struct poll_iocb *req = &aiocb->poll;

	/*
	 * Fix CVE-2021-39698
	 *  +48 lines
	 */
	if (req->file->f_op == &(*klpe_signalfd_fops)) {
		/*
		 * The wait_queue_head at req->head might not exist
		 * anymore, and it must be made sure that
		 * signalfd_cleanup() or aio_poll_wake(POLLFREE)
		 * respectively has not been invoked on it yet before
		 * accessing it.
		 *
		 * Note that once a aio poll request has been
		 * completed, either inline in aio_poll_wake() or from
		 * aio_poll_complete_work() scheduled therefrom, it
		 * will get removed from the ->active_reqs list and
		 * can't be cancelled anymore. That is, by the mere
		 * fact of being here, we know it's not been completed
		 * yet (and won't be while here because the
		 * ctx->ctx_lock is being held).
		 *
		 * Thus, the case of concern here is that
		 * aio_poll_wake(POLLFREE) has been run and set
		 * aiocb->ki_res.res to -ECANCELED in order to record
		 * the POLLFREE event, but aio_poll_complete_work()
		 * has not made it past acquisition of the
		 * ctx->ctx_lock. In particular, the -ECANCELED in
		 * aiocb->ki_res.res would not have been overwritten
		 * by aio_poll_complete_work() yet.
		 */
		if (req->cancelled)
			return 0;

		/*
		 * For signalfds, the wait_queue_head referenced by
		 * req->head is from a SLAB_TYPESAFE_BY_RCU slab cache,
		 * c.f. the comment in signalfd_cleanup() and the
		 * slab cache initializtion in proc_caches_init().
		 * If req->ki_res.res != -ECANCELED below,
		 * then aio_poll_wake(POLLFREE) hasn't run yet and
		 * the RCU read side section guarantees that req->head
		 * points to valid memory storing *some* wait_queue_head,
		 * but not necessarily the original one by the time
		 * aio_poll()->signalfd_poll() had been executed. In either
		 * case, it is perfectly fine to take req->head->lock.
		 */
		rcu_read_lock();
		if (READ_ONCE(aiocb->ki_res.res) == -ECANCELED) {
			rcu_read_unlock();
			return 0;
		}
	}

	spin_lock(&req->head->lock);

	/*
	 * Fix CVE-2021-39698
	 *  +10 lines
	 */
	/*
	 * Ok, got the wait_queue_head ->lock, now being fully
	 * serialized with aio_poll_wake(POLLFREE).
	 */
	if (req->file->f_op == &(*klpe_signalfd_fops) &&
	    aiocb->ki_res.res == -ECANCELED) {
		spin_unlock(&req->head->lock);
		rcu_read_unlock();
		return 0;
	}

	WRITE_ONCE(req->cancelled, true);
	if (!list_empty(&req->wait.entry)) {
		list_del_init(&req->wait.entry);
		schedule_work(&aiocb->poll.work);
	}
	spin_unlock(&req->head->lock);

	/*
	 * Fix CVE-2021-39698
	 *  +2 lines
	 */
	if (req->file->f_op == &(*klpe_signalfd_fops))
		rcu_read_unlock();

	return 0;
}

int klpp_aio_poll_wake(struct wait_queue_entry *wait, unsigned mode, int sync,
		void *key)
{
	struct poll_iocb *req = container_of(wait, struct poll_iocb, wait);
	struct aio_kiocb *iocb = container_of(req, struct aio_kiocb, poll);
	__poll_t mask = key_to_poll(key);
	unsigned long flags;

	/* for instances that support it check for an event match first: */
	if (mask && !(mask & req->events))
		return 0;

	/*
	 * Fix CVE-2021-39698
	 *  +10 lines
	 */
	if (!mask && req->file->f_op == &(*klpe_signalfd_fops)) {
		/*
		 * Anything but a POLLFREE/EPOLLHUP event. Don't even
		 * bother to schedule the work below, signalfd_poll()
		 * won't return anything in the context of the worker
		 * thread anyway. Leave us on the wqh so that future
		 * POLLFREE events won't get missed.
		 */
		return 0;
	}

	list_del_init(&req->wait.entry);

	if (mask && spin_trylock_irqsave(&iocb->ki_ctx->ctx_lock, flags)) {
		struct kioctx *ctx = iocb->ki_ctx;

		/*
		 * Try to complete the iocb inline if we can. Use
		 * irqsave/irqrestore because not all filesystems (e.g. fuse)
		 * call this function with IRQs disabled and because IRQs
		 * have to be disabled before ctx_lock is obtained.
		 */
		list_del(&iocb->ki_list);
		iocb->ki_res.res = mangle_poll(mask);
		req->done = true;
		if (iocb->ki_eventfd && klpr_eventfd_signal_count()) {
			iocb = NULL;
			INIT_WORK(&req->work, (*klpe_aio_poll_put_work));
			schedule_work(&req->work);
		}
		spin_unlock_irqrestore(&ctx->ctx_lock, flags);
		if (iocb)
			klpr_iocb_put(iocb);
	/*
	 * Fix CVE-2021-39698
	 *  +31 lines
	 */
	} else if (req->file->f_op == &(*klpe_signalfd_fops)) {
		/*
		 * This is a POLLFREE/EPOLLHUP event from
		 * signalfd_cleanup() and the request could not get
		 * completed inline, because the spin_trylock() above
		 * failed to acquire the required ctx lock. Note that
		 * before the livepatch, the scheduled
		 * aio_poll_complete_work() would have failed to
		 * extract the event via signalfd_poll() from the
		 * context of the worker thread and a UAF would have
		 * occured upon the readdition to the work_queue_head.
		 * However, simply skipping the work scheduling below
		 * alltogether to "fix" the UAF in a kind of bug
		 * compatible way won't work: aio_poll_cancel()
		 * expects the work to have been scheduled if
		 * list_empty(&req->wait.entry) is found to be true.
		 * One approach would be to mimic cancellation,
		 * i.e. to set req->cancelled and to schedule the
		 * completion work. However, userspace might not
		 * expect this: it would look like a successful
		 * completion with an empty poll result.  Instead, do
		 * better and signal to a livepatched
		 * aio_poll_complete_work() via req->ki_res.res that
		 * there had been a POLLFREE/EPOLLHUP event and let it
		 * handle this properly. Note that a not livepatched
		 * aio_poll_complete_work() would simply ignore this
		 * field and potentially reset it in the course of
		 * cancellation, if any.
		 */
		WRITE_ONCE(iocb->ki_res.res, -ECANCELED);
		schedule_work(&req->work);
	} else {
		schedule_work(&req->work);
	}
	return 1;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1196959.h"
#include "bsc1196959_common.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "aio_poll_put_work", (void *)&klpe_aio_poll_put_work },
	{ "eventfd_wake_count", (void *)&klpe_eventfd_wake_count },
	{ "kiocb_cachep", (void *)&klpe_kiocb_cachep },
	{ "refill_reqs_available", (void *)&klpe_refill_reqs_available },
	{ "signalfd_fops", (void *)&klpe_signalfd_fops },
};

int livepatch_bsc1196959_aio_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
